# -*- coding: utf-8 -*-

from typing import Any, Optional, Union
import aiohttp
import orjson
import time

__all__ = ()

OSU_API_BASE = 'https://osu.ppy.sh/api/v2'

# the OsuAPIWrapper classes functions will accept both
# string and integral representations of gamemodes.
API_GameMode = Union[str, int]

class OsuAPIWrapper:
    def __init__(self, client_id: int, client_secret: str) -> None:
        self.client_id = client_id
        self.client_secret = client_secret

        self.access_token = {
            'token': None,
            'timeout': 0
        }

    async def __aenter__(self):
        self.http_sess = aiohttp.ClientSession(json_serialize=orjson.dumps)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.http_sess.close()

    async def _authorize(self) -> None:
        """Request authorization from the osu!api."""
        url = 'https://osu.ppy.sh/oauth/token'
        params = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'client_credentials',
            'scope': 'public'
        }

        async with self.http_sess.post(url, data=params) as resp:
            json = await resp.json()

            self.access_token = {
                'token': json['access_token'],
                'timeout': time.time() + json['expires_in']
            }

    async def _request(self, url: str, params: Optional[dict[str, Any]] = None) -> None:
        """Perform a request to the osu!api."""
        # check if oauth2 token is expired
        if time.time() > self.access_token['timeout']:
            await self._authorize()

        headers = {'Authorization': f'Bearer {self.access_token["token"]}'}

        # XXX: i don't think i need to support POST? will if needed
        async with self.http_sess.get(url, params=params, headers=headers) as resp:
            json = await resp.json()

        return json

    # GET /users/{user_id}/{mode?}
    async def get_user(self, user_id: int, mode: Optional[API_GameMode] = None) -> None:
        """This endpoint returns the detail of specified user."""
        url = f'{OSU_API_BASE}/users/{user_id}'

        # the user may pass a gamemode as either string or int,
        # but the osu!api requires the string version.
        if mode is not None:
            if isinstance(mode, str):
                if mode not in ('osu', 'taiko', 'fruits', 'mania'):
                    return

            elif isinstance(mode, int):
                if not (0 < mode <= 3):
                    return

                mode = ('osu', 'taiko', 'fruits', 'mania')[mode]

            url += f'/{mode}'

        return await self._request(url)


    # [TODO] GET /users/{user_id}/scores/{type}

    # TODO: finish lol



# -*- coding: utf-8 -*-

from collections import namedtuple
from functools import wraps
from typing import Callable
from enum import IntEnum
from time import time_ns
from string import ascii_letters, digits
from random import choice
from datetime import (
    datetime as dt,
    timezone as tz,
    timedelta as td
)

__all__ = ('get_timestamp', '_isdecimal', 'rstring',
           'async_cache', 'TimeScale', 'timef')

ts_fmt = ('%I:%M:%S%p', '%d/%m/%Y %I:%M:%S%p')
tz_est = tz(td(hours = -5), 'EST') # TODO: fix for edt switch lol
def get_timestamp(full: bool = False) -> str:
    return f'{dt.now(tz = tz_est):{ts_fmt[full]}}'

def _isdecimal(s: str, _float: bool = False,
               _negative: bool = False) -> None:
    if _float:
        s = s.replace('.', '', 1)

    if _negative:
        s = s.replace('-', '', 1)

    return s.isdecimal()

__chars = ascii_letters + digits
def rstring(l: int, seq: str = __chars) -> str:
    return ''.join((choice(seq) for _ in range(l)))

_CacheStats = namedtuple('CacheStats', ['hits', 'misses'])

def _make_key(args, kwargs):
    """Get a hashed value out of args & kwargs."""
    key = args

    if kwargs:
        for item in kwargs.items():
            key += item

    return hash(key)

def async_cache(f):
    """Lightweight unbounded async function cache."""
    cache = {}
    hits = misses = 0

    @wraps(f)
    async def wrapper(*args, **kwargs):
        key = _make_key(args, kwargs)

        if key in cache:
            nonlocal hits
            hits += 1

            val = cache[key]
        else:
            nonlocal misses
            misses += 1

            val = await f(*args, **kwargs)
            cache[key] = val

        return val

    def cache_clear():
        nonlocal hits, misses
        hits = misses = 0
        cache.clear()

    def cache_stats():
        nonlocal hits, misses
        return _CacheStats(hits, misses)

    wrapper.cache_clear = cache_clear
    wrapper.cache_stats = cache_stats
    return wrapper

class TimeScale(IntEnum):
    Years = 0
    Months = 1
    Days = 2
    Hours = 3
    Minutes = 4
    Seconds = 5
    Milliseconds = 6
    Microseconds = 7
    Nanoseconds = 8

Call = namedtuple('CallEvent', ['duration', 'time',
                                'args', 'kwargs'])

timescale_divisor = lambda ts: { # lol
    TimeScale.Years: 1000 * 1000 * 1000 * 60 * 60 * 24 * 365,
    TimeScale.Months: 1000 * 1000 * 1000 * 60 * 60 * 24 * 30,
    TimeScale.Days: 1000 * 1000 * 1000 * 60 * 60 * 24,
    TimeScale.Hours: 1000 * 1000 * 1000 * 60 * 60,
    TimeScale.Minutes: 1000 * 1000 * 1000 * 60,
    TimeScale.Seconds: 1000 * 1000 * 1000,
    TimeScale.Milliseconds: 1000 * 1000,
    TimeScale.Microseconds: 1000,
    TimeScale.Nanoseconds: 1
}[ts]

def timef(f: Callable):
    calls: list[Call] = []

    @wraps(f)
    def wrapper(*args, **kwargs):
        # all timing is done in nanoseconds,
        # conversion will be done later when
        # displaying data.
        start = time_ns() # start clock
        result = f(*args, **kwargs) # run func
        end = time_ns() # stop clock

        duration = end - start

        # add to times log
        nonlocal calls
        calls.append(Call(duration, end, args, kwargs))

        return result

    def average(ts: TimeScale = TimeScale.Milliseconds):
        nonlocal calls
        total = sum([c.duration for c in calls])
        average = total / len(calls)
        return average / timescale_divisor(ts)

    wrapper.calls = calls # maybe temp?
    wrapper.average = average
    return wrapper



import setuptools
import re

with open('requirements.txt', 'r') as f:
    requirements = f.read().splitlines()

with open('cmyui/__init__.py', 'r') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name = 'cmyui',
    author = 'cmyui',
    url = 'https://github.com/cmyui/cmyui_pkg',
    version = version,
    packages = setuptools.find_packages(),
    description = 'Tools I find myself constantly rebuilding and reusing.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    license = 'MIT',
    install_requires = requirements,
    python_requires = '>=3.9',
    package_data = {
        'cmyui': ['py.typed'],
    },
    classifiers = [
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
    ]
)
